"""Initial schema

Revision ID: 1eab976dea29
Revises:
Create Date: 2023-08-02 19:05:29.656326

"""
from alembic import op
import sqlalchemy as sa
from pgvector.sqlalchemy import Vector
import uuid

# revision identifiers, used by Alembic.
revision = '1eab976dea29'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:

    conn = op.get_bind()
    conn.execute(sa.text('CREATE EXTENSION IF NOT EXISTS vector'))

    op.create_table(
        'langchain_pg_collection',
        sa.Column('uuid', sa.UUID(), primary_key=True, default=uuid.uuid4),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('cmetadata', sa.JSON(), nullable=True),
    )

    op.create_table(
        'langchain_pg_embedding',
        sa.Column('uuid', sa.UUID(), primary_key=True, default=uuid.uuid4),
        sa.Column('collection_id', sa.UUID()),
        sa.Column('embedding', Vector()),
        sa.Column('document', sa.String(), nullable=True),
        sa.Column('cmetadata', sa.JSON(), nullable=True),
        sa.Column('custom_id', sa.String(), nullable=True),
    )
    op.create_foreign_key(
        constraint_name='langchain_pg_embedding_collection_id_fkey',
        source_table='langchain_pg_embedding',
        referent_table='langchain_pg_collection',
        local_cols=['collection_id'],
        remote_cols=['uuid'],
        onupdate='NO ACTION',
        ondelete='CASCADE',
    )

    op.create_table(
        'chat_history',
        sa.Column('uuid', sa.UUID(), primary_key=True, default=uuid.uuid4),
        sa.Column('session_id', sa.UUID(), nullable=True),
        sa.Column('collection_id', sa.UUID()),
        sa.Column('question', sa.String()),
        sa.Column('answer', sa.String()),
        sa.Column(
            'created',
            sa.TIMESTAMP(timezone=False),
            server_default=sa.func.current_timestamp()
        ),
        sa.Column('cmetadata', sa.JSON(), nullable=True),
    )
    op.create_foreign_key(
        constraint_name='chat_history_collection_id_fkey',
        source_table='chat_history',
        referent_table='langchain_pg_collection',
        local_cols=['collection_id'],
        remote_cols=['uuid'],
        onupdate='NO ACTION',
        ondelete='CASCADE',
    )


def downgrade() -> None:
    op.drop_table('langchain_pg_embedding')
    op.drop_table('chat_history')
    op.drop_table('langchain_pg_collection')



"""Import an HTML file performing embedding and store."""
import sys
import os
from langchain.docstore.document import Document
from langchain.document_loaders import BSHTMLLoader
from dotenv import load_dotenv
from brevia import index, load_file


def load_documents(
    file_path: str,
    collection: str,
) -> list[Document]:
    """ Load documents in collection index from a CSV file """
    if not os.path.isfile(file_path):
        print(f"File {file_path} does not exist.")
        sys.exit()

    loader = BSHTMLLoader(file_path=file_path)
    documents = loader.load()

    # uncomment to use custom parsing instead of plain HTML loader
    # documents = read_html(
    #     file_path=file_path,
    #     bs_kwargs={'features': 'html.parser'},
    # )
    texts_count = 0
    for doc in documents:
        doc.page_content = load_file.cleanup_text(doc.page_content)
        texts_count += index.add_document(
            document=doc,
            collection_name=collection,
        )

    print(
        f"""Index collection '{collection}'
        updated with {len(documents)} documents and {texts_count} texts"""
    )


# def read_html(
#     file_path: str,
#     bs_kwargs: dict = None,
# ):
#     """ Read HTML file and create documents with custom parsing """
#     from bs4 import BeautifulSoup

#     with open(file_path, 'r') as f:
#         soup = BeautifulSoup(f, **bs_kwargs)

#     meta = {'source': os.path.basename(file_path)}
#     items = soup.find_all('div', {'class': 'faq-entry-container'})
#     docs = []
#     for item in items:
#         doc = Document(page_content=item.get_text(strip=True), metadata=meta)
#         docs.append(doc)

#     return docs


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print("Usage: python html_import.py <html_file> <collection>")
        sys.exit()
    load_dotenv()
    index.init_index()
    load_documents(
        file_path=sys.argv[1],
        collection=sys.argv[2]
    )



""" DB connection utility functions """
import os
import subprocess
import sys
from urllib import parse
import sqlalchemy
from sqlalchemy.orm import Session
from sqlalchemy.exc import OperationalError
from langchain.vectorstores._pgvector_data_models import CollectionStore


def connection_string() -> str:
    """ Postgres+pgvector db connection string """
    driver = os.environ.get('PGVECTOR_DRIVER')
    host = os.environ.get('PGVECTOR_HOST')
    port = int(os.environ.get('PGVECTOR_PORT'))
    database = os.environ.get('PGVECTOR_DATABASE')
    user = os.environ.get('PGVECTOR_USER')
    password = parse.quote_plus(os.environ.get('PGVECTOR_PASSWORD'))

    return f"postgresql+{driver}://{user}:{password}@{host}:{port}/{database}"


def db_connection() -> sqlalchemy.engine.Connection:
    """ SQLAlchemy db connection """
    engine = sqlalchemy.create_engine(
        connection_string(),
        pool_size=int(os.environ.get('PGVECTOR_POOL_SIZE', '10')),
    )
    return engine.connect()


def test_connection() -> bool:
    """ Test db connection with a simple query """
    try:
        with Session(db_connection()) as session:
            (
                session.query(CollectionStore.uuid)
                .limit(1)
                .all()
            )
        return True
    except OperationalError:
        print('Error performing a simple SQL query', file=sys.stderr)
        return False


def psql_command(
    cmd: str,
) -> bool:
    """Perform command on db using `psql`"""
    host = os.environ.get('PGVECTOR_HOST')
    port = os.environ.get('PGVECTOR_PORT')
    database = os.environ.get('PGVECTOR_DATABASE')
    user = os.environ.get('PGVECTOR_USER')
    password = os.environ.get('PGVECTOR_PASSWORD')

    cmd = ['psql', '-U', user, '-h', host, '-p', port, '-c', f"{cmd}", database]

    with subprocess.Popen(
        cmd,
        env={'PGPASSWORD': password},
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE
    ) as proc:
        res = proc.communicate()
        output = res[0].decode('UTF-8')
        print(f"psql output: {output}")
        if proc.returncode != 0:
            print('Error launching psql command', file=sys.stderr)
            return False

    print('Psql command executed successfully')
    return True



"""Chat history table & utilities"""
from typing import List
import uuid
import logging
from datetime import datetime
from os import environ
from langchain.vectorstores.pgvector import BaseModel
from langchain.vectorstores._pgvector_data_models import CollectionStore
import sqlalchemy
from sqlalchemy.dialects.postgresql import JSON, UUID
from sqlalchemy.orm import Mapped, Query, Session
from sqlalchemy.sql.expression import BinaryExpression
from brevia.connection import db_connection
from brevia.index import get_embeddings


class ChatHistoryStore(BaseModel):
    # pylint: disable=too-few-public-methods
    """ Chat History table """
    __tablename__ = "chat_history"

    session_id = sqlalchemy.Column(UUID(as_uuid=True))
    collection_id: Mapped[UUID] = sqlalchemy.Column(
        UUID(as_uuid=True),
        sqlalchemy.ForeignKey(
            f"{CollectionStore.__tablename__}.uuid",
            ondelete="CASCADE",
        ),
    )
    question = sqlalchemy.Column(sqlalchemy.String)
    answer = sqlalchemy.Column(sqlalchemy.String)
    # pylint: disable=not-callable
    created = sqlalchemy.Column(
        sqlalchemy.DateTime(timezone=False),
        nullable=False,
        server_default=sqlalchemy.sql.func.now()
    )
    cmetadata = sqlalchemy.Column(JSON, nullable=True)


def history(chat_history: list, session: str = None):
    """ Load chat history from input or from DB """
    if len(chat_history) > 0:
        return [(x["query"], x["answer"]) for x in chat_history]

    if not is_valid_uuid(session) or bool(environ.get('QA_NO_CHAT_HISTORY')):
        return []

    return history_from_db(session)


def is_related(chat_history: list, question: str):
    """
    Determina se una domanda è correlata a una sequenza di frasi.
    Utilizza gli embedding delle frasi e della domanda per calcolare il prodotto
    scalare tra i vettori (in questo caso = similarity) e confrontarlo
    con una soglia specificata dalle variabili d'ambiente.
    """
    embeddings = get_embeddings()
    q_e = embeddings.embed_query(question)
    h_e = embeddings.embed_query(
        ''.join([sentence for tuple in chat_history for sentence in tuple])
    )
    sim = dot_product(q_e, h_e)
    logging.getLogger(__name__).info("similarity: %s", sim)
    threshold = float(environ.get('QA_FOLLOWUP_SIM_THRESHOLD', False))
    return sim >= threshold


def dot_product(v1_list, v2_list):
    """
    Calcola il prodotto scalare tra due vettori (similarity).
    """
    return sum(x * y for x, y in zip(v1_list, v2_list))


def history_from_db(session_id: str) -> List[tuple[str, str]]:
    """ Load chat history from DB """
    connection = db_connection()
    ChatHistoryStore.metadata.create_all(connection)

    filter_by = ChatHistoryStore.session_id == session_id
    with Session(connection) as session:
        results: List[ChatHistoryStore] = (
            session.query(ChatHistoryStore)
            .filter(filter_by)
            .order_by(sqlalchemy.desc(ChatHistoryStore.created))
            .limit(3)
            .all()
        )

    chat_hist = [(x.question, x.answer) for x in results]
    chat_hist.reverse()

    return chat_hist


def add_history(
    session_id: str,
    collection: str,
    question: str,
    answer: str,
    metadata: dict | None = None,
):
    """Save chat history item to database """
    if not is_valid_uuid(session_id):
        return

    with Session(db_connection()) as session:
        collection_store = CollectionStore.get_by_name(session, collection)
        if not collection_store:
            raise ValueError("Collection not found")
        chat_history_store = ChatHistoryStore(
            session_id=session_id,
            collection_id=collection_store.uuid,
            question=question,
            answer=answer,
            cmetadata=metadata,
        )
        session.add(chat_history_store)
        session.commit()


def is_valid_uuid(val) -> bool:
    """ Check UUID validity """
    try:
        uuid.UUID(str(val))
        return True
    except (ValueError, TypeError):
        return False


def get_history(
    max_date: str | None = None,
    collection: str | None = None,
    page: int = 1,
    page_size: int = 50,
):
    """
        Read chat history with optional date and collection filters
        using pagination data in response
    """
    max_date = datetime.now() if max_date is None else max_date
    filter_date = ChatHistoryStore.created <= max_date
    filter_collection = CollectionStore.name == collection
    if collection is None:
        filter_collection = CollectionStore.name is not None

    page = max(1, page)  # min page number is 1
    page_size = min(1000, page_size)  # max page size is 1000

    query = get_history_query(
        filter_date=filter_date,
        filter_collection=filter_collection,
    )
    count = query.count()
    offset = (page - 1) * page_size
    results = [u._asdict() for u in query.offset(offset).limit(page_size).all()]

    pcount = int(count / page_size)
    pcount += 0 if (count % page_size) == 0 else 1

    return {
        'data': results,
        'meta': {
            'pagination': {
                'count': count,
                'page': page,
                'page_count': pcount,
                'page_items': len(results),
                'page_size': page_size,
            },
        }
    }


def get_history_query(
    filter_date: BinaryExpression,
    filter_collection: BinaryExpression,
) -> Query:
    """Return get history query"""
    with Session(db_connection()) as session:
        return (
            session.query(
                ChatHistoryStore.question,
                ChatHistoryStore.answer,
                ChatHistoryStore.session_id,
                ChatHistoryStore.cmetadata,
                ChatHistoryStore.created,
                CollectionStore.name.label('collection'),
            )
            .join(
                CollectionStore,
                CollectionStore.uuid == ChatHistoryStore.collection_id
            )
            .filter(filter_date, filter_collection)
            .order_by(sqlalchemy.desc(ChatHistoryStore.created))
        )



"""API endpoints definitions to handle audio input"""
from fastapi import APIRouter
from brevia.dependencies import get_dependencies
from brevia import chat_history

router = APIRouter()


@router.get('/chat_history', dependencies=get_dependencies(json_content_type=False))
def read_chat_history(
    max_date: str | None = None,
    collection: str | None = None,
    page: int = 1,
    page_size: int = 50,
):
    """ /chat_history endpoint, read stored chat history """
    return chat_history.get_history(
        max_date=max_date,
        collection=collection,
        page=page,
        page_size=page_size,
    )



"""Import a PDF file performing embedding and store."""
import sys
from dotenv import load_dotenv
from brevia import index


def load_documents():
    """Load documents from PDF using command line args"""
    load_dotenv()
    index.init_index()
    num_docs = index.load_pdf_file(
        file_path=sys.argv[1],
        collection_name=sys.argv[2],
        page_from=None if len(sys.argv) < 4 else int(sys.argv[3]),
        page_to=None if len(sys.argv) < 5 else int(sys.argv[4]),
    )

    print(
        f"""
        Index collection '{sys.argv[2]}'
        updated with {num_docs} documents."""
    )


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print("Usage: python pdf_import.py <pdf_file_name> "
              "<collection_name> [page_from] [page_to]")
        sys.exit()
    load_documents()



"""Returning text summarize or question-answering chain against a vector database."""
from os import environ, path
import logging
from langchain.docstore.document import Document
from langchain.vectorstores.pgvector import PGVector, DistanceStrategy
from langchain.vectorstores._pgvector_data_models import CollectionStore
from langchain.callbacks.base import BaseCallbackHandler
from langchain.chains.base import Chain
from langchain.chains.summarize import load_summarize_chain
from langchain.text_splitter import TokenTextSplitter
from langchain.chains import ConversationalRetrievalChain
from langchain.chains.question_answering import load_qa_chain
from langchain.chains.llm import LLMChain
from langchain import OpenAI
from langchain.prompts import load_prompt
from langchain.prompts import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)
from langchain.prompts.loading import load_prompt_from_config
from langchain.chat_models import ChatOpenAI
from brevia import connection, index
from brevia.callback import AsyncLoggingCallbackHandler, LoggingCallbackHandler


def load_brevia_prompt(prompts: dict | None) -> ChatPromptTemplate:
    """ load prompts for Q/A functions """

    prompts_path = f'{path.dirname(__file__)}/prompts'
    system = load_prompt(f'{prompts_path}/qa/default.system.yaml')
    human = load_prompt(f'{prompts_path}/qa/default.human.yaml')

    if prompts:
        if prompts.get('system'):
            system = load_prompt_from_config(prompts.get('system'))
        if prompts.get('human'):
            human = load_prompt_from_config(prompts.get('human'))

    system_message_prompt = SystemMessagePromptTemplate(prompt=system)
    human_message_prompt = HumanMessagePromptTemplate(prompt=human)

    return ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )


def load_condense_prompt(prompts: dict | None) -> ChatPromptTemplate:
    """
        Check if specific few-shot prompt file exists for the collection, if not,
        check for condense prompt file, otherwise, load default condense prompt file.
    """
    if prompts:
        if prompts.get('few'):
            return load_prompt_from_config(prompts.get('few'))
        if prompts.get('condense'):
            return load_prompt_from_config(prompts.get('condense'))

    prompts_path = f'{path.dirname(__file__)}/prompts'

    return load_prompt(f'{prompts_path}/qa/default.condense.yaml')


def search_vector_qa(
    query: str,
    collection: str,
    docs_num: int = int(environ.get('SEARCH_DOCS_NUM', 4)),
    distance_strategy_name: str = 'cosine',
) -> list[tuple[Document, float]]:
    """ Perform a similarity search on vector index """

    docsearch = PGVector(
        connection_string=connection.connection_string(),
        embedding_function=index.get_embeddings(),
        collection_name=collection,
        distance_strategy=distance_strategy(distance_strategy_name),
    )

    return docsearch.similarity_search_with_score(query, k=docs_num)


def distance_strategy(strategy: str):
    """Distance strategy from type name"""
    distance_map = {
        'euclidean': DistanceStrategy.EUCLIDEAN,
        'cosine': DistanceStrategy.COSINE,
        'max': DistanceStrategy.MAX_INNER_PRODUCT,
    }
    if strategy not in distance_map:
        return DistanceStrategy.COSINE

    return distance_map[strategy]


def conversation_chain(
    # pylint: disable=too-many-arguments
    collection: CollectionStore,
    docs_num: int | None,
    source_docs: bool = True,
    distance_strategy_name: str = 'cosine',
    streaming: bool = False,
    answer_callbacks: list[BaseCallbackHandler] = [],
    conversation_callbacks: list[BaseCallbackHandler] = [],
) -> Chain:
    """
        Return conversation chain for Q/A with embdedded dataset knowledge

        collection: name of collection to questioning
        source_docs: retrive source docs
        docs_num: number of docs to retrieve to create context
        distance_strategy_name: distance strategy to use
        streaming: activate streaming (default False),
        answer_callbacks: callbacks to use in the final LLM answer to enable streaming
        conversation_callbacks: callback to handle conversation results

        can implement "vectordbkwargs" into quest_dict:
            {
                "search_distance": 0.9
            }
    """
    if docs_num is None:
        docs_num = int(environ.get('SEARCH_DOCS_NUM', 4))

    docsearch = PGVector(
        connection_string=connection.connection_string(),
        embedding_function=index.get_embeddings(),
        collection_name=collection.name,
        distance_strategy=distance_strategy(distance_strategy_name),
    )

    prompts = collection.cmetadata.get('prompts')

    # LLM to rewrite follow-up question
    fup_llm = OpenAI(
        model_name=environ.get('QA_FOLLOWUP_MODEL'),
        temperature=float(environ.get('QA_FOLLOWUP_TEMPERATURE', 0)),
        max_tokens=int(environ.get('QA_FOLLOWUP_MAX_TOKENS', 200)),
        verbose=environ.get('VERBOSE_MODE', False),
    )

    logging_handler = AsyncLoggingCallbackHandler()
    # Rewrite question using chat history (if any)
    question_generator = LLMChain(
        llm=fup_llm,
        prompt=load_condense_prompt(prompts),
        verbose=environ.get('VERBOSE_MODE', False),
        callbacks=[logging_handler],
    )

    # LLM to use in final prompt
    answer_callbacks.append(logging_handler)
    chatllm = ChatOpenAI(
        model_name=environ.get('QA_COMPLETIONS_MODEL'),
        temperature=float(environ.get('QA_TEMPERATURE', 0)),
        max_tokens=int(environ.get('QA_MAX_TOKENS', 800)),
        callbacks=answer_callbacks,
        streaming=streaming,
        verbose=environ.get('VERBOSE_MODE', False),
    )

    # this chain use "stuff" to elaborate context
    doc_chain = load_qa_chain(
        llm=chatllm,
        prompt=load_brevia_prompt(prompts),
        chain_type="stuff",
        verbose=environ.get('VERBOSE_MODE', False),
        callbacks=[logging_handler],
    )

    # main chain, do all the jobs
    search_kwargs = {'k': docs_num}

    conversation_callbacks.append(logging_handler)
    return ConversationalRetrievalChain(
        retriever=docsearch.as_retriever(search_kwargs=search_kwargs),
        combine_docs_chain=doc_chain,
        return_source_documents=source_docs,
        question_generator=question_generator,
        callbacks=conversation_callbacks,
        verbose=environ.get('VERBOSE_MODE', False)
    )


def summarize(
    text: str,
    num_items: int = int(environ.get('SUMM_NUM_ITEMS', 5)),
    summ_prompt: str = environ.get('SUMM_DEFAULT_PROMPT')
) -> str:
    """ Perform summarizing for a given text """

    if bool(environ.get('FAKE_SUMMARY')):
        print('Using FAKE summary - text truncate!!')
        logging.getLogger(__name__).warning('Using FAKE summary - text truncate!!')
        import time  # pylint: disable=import-outside-toplevel
        time.sleep(30)
        return text[:min(100, len(text)-1)]

    text_splitter = TokenTextSplitter(
        chunk_size=int(environ.get("SUMM_TOKEN_SPLITTER", 3000)),
        chunk_overlap=int(environ.get("SUMM_TOKEN_OVERLAP", 0))
    )
    texts = text_splitter.split_text(text)
    docs = [Document(page_content=t) for t in texts]
    lang = environ.get("PROMPT_LANG", 'it')
    if summ_prompt not in ['summarize', 'summarize_point', 'classificate']:
        summ_prompt = 'summarize'
    prompts_path = f'{path.dirname(__file__)}/prompts'
    prompt = load_prompt(f'{prompts_path}/summarize/yaml/{lang}.{summ_prompt}.yaml')
    logging_handler = LoggingCallbackHandler()
    chain = load_summarize_chain(
        OpenAI(
            model_name=environ.get("SUMM_COMPLETIONS_MODEL"),
            temperature=float(environ.get("SUMM_TEMPERATURE", 0)),
            max_tokens=int(environ.get("SUMM_MAX_TOKENS", 800)),
            callbacks=[logging_handler],
        ),
        chain_type='map_reduce',
        map_prompt=prompt,
        verbose=environ.get("VERBOSE_MODE", False),
        combine_prompt=prompt,
        callbacks=[logging_handler],
    )
    return chain.run(**{'input_documents': docs, 'num_items': num_items})



"""API endpoints definitions to handle audio input"""
from fastapi import APIRouter, Response, status
from brevia.dependencies import get_dependencies
from brevia import connection

router = APIRouter()


@router.api_route(
    '/status',
    methods=['GET', 'HEAD'],
    dependencies=get_dependencies(json_content_type=False),
)
def api_status(response: Response):
    """ /status endpoint, safety check """
    db_status = connection.test_connection()
    if not db_status:
        response.status_code = status.HTTP_503_SERVICE_UNAVAILABLE

    return {
        'db_status': 'OK' if db_status else 'KO'
    }



"""Utility commands for applications"""
import click
from dotenv import load_dotenv
from brevia import alembic


@click.command()
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbose mode")
def db_current_cmd(verbose):
    """Display current database revision"""
    load_dotenv()
    alembic.current(verbose)


@click.command()
@click.option("-r", "--revision", default="head", help="Revision target")
def db_upgrade_cmd(revision):
    """Upgrade to a later database revision"""
    load_dotenv()
    alembic.upgrade(revision)


@click.command()
@click.option("-r", "--revision", required=True, help="Revision target")
def db_downgrade_cmd(revision):
    """Revert to a previous database revision"""
    load_dotenv()
    alembic.downgrade(revision)



from logging import config, getLogger
from os import path, getcwd
import sys
import click
from dotenv import load_dotenv
import yaml
from brevia.async_jobs import create_service


# A `test_service.yml` in the root folder is needed.
# `service` and `payload` keys are required with this structure:
#
# service: brevia.extensions.my_extension.services.MyService
# output_path: path/to/output.txt
# payload:
#   file_path: /path/to/file
#   param1: value1
#   param2: value2

@click.command()
@click.option("-n", "--num", default=1, help="Number of attempts")
@click.option(
    "-f",
    "--file-path",
    default=f'{getcwd()}/test_service.yml',
    help="yaml file path"
)
def test_service(num: int = 1, file_path: str = f'{getcwd()}/test_service.yml'):
    """Service test"""
    load_dotenv()
    # allow module import from current working directory
    sys.path.append(getcwd())
    # initialize logging from optional log.ini
    log_ini_path = f'{getcwd()}/log.ini'
    if path.exists(log_ini_path):
        config.fileConfig(log_ini_path)

    if not path.isfile(file_path):
        raise FileNotFoundError(file_path)

    with open(file_path, 'r') as stream:
        parsed_yaml = yaml.safe_load(stream)

    attempt = 1
    while attempt <= num:
        test_run_service(
            service=parsed_yaml.get('service'),
            payload=parsed_yaml.get('payload'),
            output_path=parsed_yaml.get('output_path'),
            num_attempt=attempt,
        )
        attempt += 1


def test_save_result(file_path: str, num_attempt: int, output: str):
    """Save JOB result output to file, adding num attempt to file name"""
    file_name, file_extension = path.splitext(file_path)

    with open(f'{file_name}-{num_attempt}{file_extension}', 'w') as stream:
        stream.write(output)


def test_run_service(
    service: str,
    payload: dict,
    output_path: str = None,
    num_attempt: int = 1
):
    """Create and run service"""
    service = create_service(service)
    log = getLogger(__file__)
    attempt_message = f'Attempt: {num_attempt} - File: {payload.get("file_path")}'
    log.info('%s - start', attempt_message)
    result = service.run(payload)
    log.info('%s - end', attempt_message)

    if output_path is None:
        log.info('%s - `output_path` not specified, file not saved', attempt_message)
    else:
        test_save_result(
            file_path=output_path,
            num_attempt=num_attempt,
            output=result.get('output')
        )



"""Import a CSV file performing embedding and store."""
import sys
import os
from langchain.docstore.document import Document
from langchain.document_loaders.csv_loader import CSVLoader
from dotenv import load_dotenv
from brevia import index, load_file


def load_documents(
    file_path: str,
    collection: str,
) -> list[Document]:
    """ Load documents in collection index from a CSV file """
    if not os.path.isfile(file_path):
        print(f"File {file_path} does not exist.")
        sys.exit()

    loader = CSVLoader(
        file_path=file_path,
        csv_args={
            'delimiter': ',',
            'quotechar': '"',
            'doublequote': True,
            'skipinitialspace': False,
        }
    )
    documents = loader.load()
    texts_count = 0
    for doc in documents:
        doc.page_content = load_file.cleanup_text(doc.page_content)
        texts_count += index.add_document(
            document=doc,
            collection_name=collection,
        )

    print(
        f"""Index collection '{collection}'
        updated with {len(documents)} documents and {texts_count} texts"""
    )


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print("Usage: python csv_import.py <csv_file_name> <collection_name>")
        sys.exit()
    load_dotenv()
    index.init_index()
    load_documents(
        file_path=sys.argv[1],
        collection=sys.argv[2]
    )



"""API endpoints definitions to handle Collections"""
from fastapi import APIRouter
from pydantic import BaseModel
from brevia.dependencies import (
    get_dependencies,
    check_collection_name_absent,
    check_collection_uuid
)
from brevia import collections

router = APIRouter()


@router.get('/collections', dependencies=get_dependencies(json_content_type=False))
async def collections_index(name: str | None = None):
    """ /collections endpoint, information on available collections """
    return collections.collections_info(collection=name)


@router.get(
    '/collections/{uuid}',
    dependencies=get_dependencies(json_content_type=False)
)
async def read_collection(uuid: str):
    """ /collections/{uuid} endpoint"""
    check_collection_uuid(uuid)

    return collections.single_collection(uuid)


class CollectionBody(BaseModel):
    """ summarize validation """
    name: str
    cmetadata: dict


@router.post('/collections', status_code=201, dependencies=get_dependencies())
def create_collection(body: CollectionBody):
    """ POST /collections endpoint"""
    check_collection_name_absent(body.name)

    return collections.create_collection(
        name=body.name,
        cmetadata=body.cmetadata,
    )


@router.patch('/collections/{uuid}', status_code=204, dependencies=get_dependencies())
def update_collection(uuid: str, body: CollectionBody):
    """ POST /collections endpoint"""
    check_collection_uuid(uuid)
    current = collections.single_collection(uuid)
    if current.name != body.name:
        # if name is changed check that it's not in use
        check_collection_name_absent(body.name)

    collections.update_collection(uuid=uuid, name=body.name, cmetadata=body.cmetadata)


@router.delete(
    '/collections/{uuid}',
    status_code=204,
    dependencies=get_dependencies(json_content_type=False)
)
def delete_collection(uuid: str):
    """ POST /collections endpoint"""
    check_collection_uuid(uuid)
    collections.delete_collection(uuid)



"""API endpoints definitions to handle audio input"""
import logging
from os import environ
from typing import Annotated
from fastapi import APIRouter, UploadFile, Form
import openai
from brevia.dependencies import (
    get_dependencies,
    save_upload_file_tmp,
)

router = APIRouter()


@router.post(
    '/transcribe',
    dependencies=get_dependencies(json_content_type=False)
)
def audio_transcriptions(
    file: UploadFile,
    language: Annotated[str, Form()],
):
    """ /transcribe endpoint, audio file transcription """
    log = logging.getLogger(__name__)
    log.info(f"Uploaded '{file.filename}' - {file.content_type} - {file.size}")
    log.info(f"Language '{language}")
    tmp_file_path = save_upload_file_tmp(file)

    with open(tmp_file_path, 'rb') as audio_file:

        result = openai.Audio.transcribe(
                model='whisper-1',
                file=audio_file,
                api_key=environ.get('OPENAI_API_KEY'),
                params={'language': language},
        )
        log.info('Audio transcription completed')
        log.info(result)
        return result



"""API endpoints definitions to handle document analysis like summarization"""
from typing import Annotated
from os import environ
from base64 import b64decode
from pathlib import Path
import tempfile
import json
import logging
from pydantic import BaseModel
from fastapi import (
    APIRouter,
    BackgroundTasks,
    HTTPException,
    status,
    Request,
    UploadFile,
    Form,
)
from brevia.dependencies import get_dependencies, save_upload_file_tmp
from brevia import async_jobs, load_file
from brevia.services import SummarizeTextService

router = APIRouter()


class SummarizeBody(BaseModel):
    """ summarize input """
    text: str
    summ_prompt: str | None = None
    num_items: int | None = 5
    token_data: bool = False


@router.post('/summarize', dependencies=get_dependencies())
def sum_documents(summarize: SummarizeBody):
    """
    /summarize endpoint:
    Summarize/classification a piece of text

    Specify prompt type:
        - summarize
        - summerize_point
        _ classificate

    num_items:
        - for summarize_point and classificate, specify number of items in output
    """
    summarize.text = load_file.cleanup_text(summarize.text).strip()
    if not summarize.text:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            'Empty text field',
        )

    service = SummarizeTextService()
    return service.run(summarize.dict())


@router.post(
    '/upload_summarize',
    dependencies=get_dependencies(json_content_type=False)
)
def upload_summarize(
    summ_prompt: Annotated[str, Form()],
    background_tasks: BackgroundTasks,
    file: UploadFile | None = None,
    file_content: Annotated[str, Form()] = '',
    num_items: Annotated[str, Form()] = environ.get('SUMM_NUM_ITEMS', '5'),
    token_data: Annotated[bool, Form()] = False,
):
    """
    Upload a PDF file and perform summarization
    See /summarize endpoint for `summ_prompt` and `num_items`
    arguments
    """
    if file:
        log = logging.getLogger(__name__)
        log.info(f"Uploaded '{file.filename}' - {file.content_type} - {file.size}")
        log.info(f"Summ type '{summ_prompt}' - items {num_items} - token {token_data}")
        tmp_path = save_upload_file_tmp(file)
    elif file_content:
        tmp_path = save_base64_tmp_file(file_content=file_content)
    else:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            'One of "file" or "file_content" form field is mandatory',
        )

    job = async_jobs.create_job(
        service='brevia.services.SummarizeFileService',
        payload={
            'file_path': tmp_path,
            'summ_prompt': summ_prompt,
            'num_items': int(num_items),
            'token_data': token_data,
        }
    )
    background_tasks.add_task(async_jobs.run_job_service, job.uuid)

    return {'job': job.uuid}


@router.post('/upload_analyze', dependencies=get_dependencies(json_content_type=False))
def upload_analyze(
    file: UploadFile,
    service: Annotated[str, Form()],
    background_tasks: BackgroundTasks,
    payload: Annotated[str, Form()] = '{}',
):
    """
    Upload a file and perform some analysis using a `service` class
    """
    log = logging.getLogger(__name__)
    log.info(f"Uploaded '{file.filename}' - {file.content_type} - {file.size}")
    tmp_path = save_upload_file_tmp(file)

    payload = json.loads(payload)
    payload['file_path'] = tmp_path
    job = async_jobs.create_job(
        service=service,
        payload=payload,
    )
    background_tasks.add_task(async_jobs.run_job_service, job.uuid)

    return {'job': job.uuid}


@router.post(
    '/summarize_binary',
    dependencies=get_dependencies(json_content_type=False)
)
async def summarize_binary(
    request: Request,
    background_tasks: BackgroundTasks,
):
    """
    Perform summarization of a PDF file
    passed via binary POST (application/octet-stream)
    """
    body: bytes = await request.body()
    tmp_file_path = save_binary_tmp_file(file_content=body)

    job = async_jobs.create_job(
        service='brevia.services.SummarizeFileService',
        payload={'file_path': tmp_file_path},
    )
    background_tasks.add_task(async_jobs.run_job_service, job.uuid)

    return {'job': job.uuid}


def save_binary_tmp_file(file_content: bytes) -> str:
    """ Save binary content to temp file, return path """
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        with open(tmp.name, 'wb') as file:
            file.write(file_content)

    return str(Path(tmp.name))


def save_base64_tmp_file(file_content: str) -> str:
    """ Save base64 file content to temp file, return path """

    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        content_as_bytes = str.encode(file_content)  # convert string to bytes
        content_recovered = b64decode(content_as_bytes)  # decode base64string
        with open(tmp.name, 'wb') as file:
            file.write(content_recovered)

    return str(Path(tmp.name))



"""Import a folder containing PDF files performing embedding and store."""
import glob
import sys
from dotenv import load_dotenv
from brevia import index


def load_documents():
    pathname = sys.argv[1] + '/*.pdf'
    for file in glob.glob(pathname=pathname):
        print(file)
        num_docs = index.load_pdf_file(
            file_path=file,
            collection_name=sys.argv[2],
        )

        print(
            f"""
            Index collection '{sys.argv[2]}'
            updated with {num_docs} documents from {file}."""
        )


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print("Usage: python folder_import.py <folder_path> <collection_name>")
        sys.exit()

    load_dotenv()
    index.init_index()
    load_documents()



"""JWT Utility methods"""
from os import environ
import click
from dotenv import load_dotenv
from jose import jwt, JWTError
from datetime import datetime, timedelta

ALGORITHM = 'HS256'


@click.command()
@click.option("-u", "--user", default="brevia", help="Token user name")
@click.option("-d", "--duration", default=15, help="Token duration in minutes")
def create_access_token(user: str, duration: int):
    """Create an access token """
    load_dotenv()
    to_encode = {'user': user}
    # expire time of the token
    expire = datetime.utcnow() + timedelta(minutes=duration)
    to_encode.update({'exp': expire})
    secret = environ.get('TOKENS_SECRET')
    encoded_jwt = jwt.encode(to_encode, secret, algorithm=ALGORITHM)
    print(encoded_jwt)


def verify_token(token: str):
    # try to decode the token, it will
    # raise a JWTError if the token is not correct
    secret = environ.get('TOKENS_SECRET')
    payload = jwt.decode(token, secret, algorithms=[ALGORITHM])
    # basic check: presence of `user` key, must be a non empty string
    if 'user' not in payload:
        raise JWTError('Bad payload format')
    user = payload['user']
    if type(user) != str or len(user.strip()) == 0:
        raise JWTError('Bad payload format')
    # further check: user must one of `TOKENS_USERS` env var (if set)
    valid_users = environ.get('TOKENS_USERS')
    if not valid_users:
        return
    if user not in valid_users.split(','):
        raise JWTError('Bad payload format')



"""Async Jobs table & utilities"""
import importlib
import logging
import time
from datetime import datetime
import sqlalchemy
from sqlalchemy.dialects.postgresql import JSON, TIMESTAMP, SMALLINT
from sqlalchemy.orm import Session
from langchain.vectorstores.pgvector import BaseModel
from brevia.connection import db_connection
from brevia.services import BaseService

MAX_DURATION = 120  # default max duration is 120 min / 2hr
MAX_ATTEMPTS = 1  # default max number of attempts


class AsyncJobsStore(BaseModel):
    # pylint: disable=too-few-public-methods,not-callable
    """ Async Jobs table """
    __tablename__ = "async_jobs"

    service = sqlalchemy.Column(sqlalchemy.String(), nullable=False)
    payload = sqlalchemy.Column(JSON())
    expires = sqlalchemy.Column(TIMESTAMP(timezone=False))
    created = sqlalchemy.Column(
        TIMESTAMP(timezone=False),
        nullable=False,
        server_default=sqlalchemy.func.current_timestamp(),
    )
    completed = sqlalchemy.Column(TIMESTAMP(timezone=False))
    locked_until = sqlalchemy.Column(TIMESTAMP(timezone=False))
    max_attempts = sqlalchemy.Column(
        SMALLINT(),
        nullable=False,
        server_default='1',
    )
    result = sqlalchemy.Column(JSON(), nullable=True)


def single_job(uuid: str) -> (AsyncJobsStore | None):
    """ Get single job by UUID """
    with Session(db_connection()) as session:
        return session.get(AsyncJobsStore, uuid)


def create_job(
    service: str,
    payload: dict,
) -> AsyncJobsStore:
    """ Create async job """
    max_duration = payload.get('max_duration', MAX_DURATION)  # max duration in minutes
    max_attempts = payload.get('max_attempts', MAX_ATTEMPTS)
    tstamp = int(time.time()) + (max_duration * max_attempts * 2 * 60)
    expires = datetime.fromtimestamp(tstamp)

    with Session(db_connection()) as session:
        job_store = AsyncJobsStore(
            service=service,
            payload=payload,
            expires=expires,
            max_attempts=max_attempts,
        )
        session.expire_on_commit = False
        session.add(job_store)
        session.commit()

        return job_store


def complete_job(
    uuid: str,
    result: dict,
) -> None:
    """ Complete async job with result """
    # pylint: disable=broad-exception-caught
    log = logging.getLogger(__name__)
    job_store = single_job(uuid)
    if not job_store:
        log.error("Job %s not found", uuid)
        return
    now = datetime.now()
    if job_store.expires and job_store.expires < now:
        log.warning("Job %s is expired at %s", uuid, job_store.expires)
        return

    try:
        save_job_result(job_store=job_store, result=result)
    except Exception as exc:
        msg = exc.args[0]
        log.error("Error completing job - %s", msg)
        result = {'error': msg}
        try:
            save_job_result(job_store=job_store, result=result, error=True)
        except Exception as ex2:
            log.error("Error saving job error - %s", ex2.args[0])


def save_job_result(job_store: AsyncJobsStore, result: dict, error: bool = False):
    """Save Job result"""
    with Session(db_connection()) as session:
        job_store.completed = datetime.now()
        job_store.result = result
        if error:
            job_store.max_attempts = max(job_store.max_attempts - 1, 0)
        session.add(job_store)
        session.expire_on_commit = False
        session.commit()


def create_service(service: str) -> BaseService:
    """ Create job service from string """
    module_name, class_name = service.rsplit('.', 1)
    module = importlib.import_module(module_name)
    if not hasattr(module, class_name):
        raise ValueError(f'Class "{class_name}" not found in "{module}"')
    service_class = getattr(module, class_name)
    if not issubclass(service_class, BaseService):
        raise ValueError(f'Class "{class_name}" must extend "BaseService"')

    return service_class()


def run_job_service(
    uuid: str,
) -> None:
    """ Create run job service """
    log = logging.getLogger(__name__)
    job_store = single_job(uuid)
    if not job_store:
        log.error("Job %s not found", uuid)
        return
    result = {}

    try:
        lock_job_service(job_store)
        service = create_service(job_store.service)
        result = service.run(job_store.payload)

    except Exception as exc:  # pylint: disable=broad-exception-caught
        msg = exc.args[0]
        log.error("Error in SummarizeFileService %s", msg)
        result = {'error': msg}

    finally:
        complete_job(uuid=uuid, result=result)


def lock_job_service(
    job_store: AsyncJobsStore
) -> None:
    """Lock job service"""
    if not is_job_available(job_store):
        raise RuntimeError(f'Job {job_store.uuid} is not available')
    payload = job_store.payload if job_store.payload else {}
    tstamp = int(time.time()) + (int(payload.get('max_duration', MAX_DURATION)) * 60)
    locked_until = datetime.fromtimestamp(tstamp)

    with Session(db_connection()) as session:
        job_store.locked_until = locked_until
        session.add(job_store)
        session.expire_on_commit = False
        session.commit()


def is_job_available(
    job_store: AsyncJobsStore
) -> bool:
    """Check if job is available"""
    now = datetime.now()
    if job_store.completed or (job_store.expires and job_store.expires < now):
        return False
    if job_store.locked_until and (job_store.locked_until > now):
        return False
    if job_store.max_attempts <= 0:
        return False

    return True



"""Import a collection from CSV postgres files."""
import sys
from os import path
from dotenv import load_dotenv
from brevia import connection, collections


def import_collection_data(
    folder_path: str,
    collection: str,
):
    """Import collection data using `psql`"""
    if collections.collection_name_exists(collection):
        print(f"Collection '{collection}' already exists, exiting")
        sys.exit(1)

    csv_file_collection = f"{folder_path}/{collection}-collection.csv"
    csv_file_embedding = f"{folder_path}/{collection}-embedding.csv"
    print(f"Importing from {csv_file_collection} and {csv_file_embedding}")
    if not path.exists(csv_file_collection):
        print(f"CSV file {csv_file_collection} not found, exiting...")
        return
    if not path.exists(csv_file_embedding):
        print(f"CSV file {csv_file_embedding} not found, exiting...")
        return

    table = 'langchain_pg_collection'
    copy = f"\\copy {table} FROM '{csv_file_collection}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    table = 'langchain_pg_embedding'
    copy = f"\\copy {table} FROM '{csv_file_embedding}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    print(f"Collection '{collection}' imported")


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print("Usage: python import_collection.py <folder_path> <collection_name>")
        sys.exit(1)
    load_dotenv()
    import_collection_data(
        folder_path=sys.argv[1],
        collection=sys.argv[2],
    )



"""Export a collection to CSV postgres files."""
import sys
from os import path
from dotenv import load_dotenv
from brevia import connection, collections


def export_collection_data(
    folder_path: str,
    collection: str,
):
    """Export collection data using `psql`"""
    if not collections.collection_name_exists(collection):
        print(f"Collection '{collection}' was not found, exiting")
        sys.exit(1)

    csv_file_collection = f"{folder_path}/{collection}-collection.csv"
    csv_file_embedding = f"{folder_path}/{collection}-embedding.csv"
    if path.exists(csv_file_collection):
        print(f"CSV file {csv_file_collection} exists, exiting...")
        return
    if path.exists(csv_file_embedding):
        print(f"CSV file {csv_file_embedding} exists, exiting...")
        return

    select = f"SELECT * FROM langchain_pg_collection WHERE name = '{collection}'"
    copy = f"\\copy ({select}) TO '{csv_file_collection}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    print(f"Collection record '{collection}' exported to '{csv_file_collection}'")

    sub = f"SELECT uuid from langchain_pg_collection WHERE name = '{collection}'"
    select = f"SELECT * FROM langchain_pg_embedding WHERE collection_id = ({sub})"
    copy = f"\\copy ({select}) TO '{csv_file_embedding}' WITH DELIMITER ',' CSV HEADER"
    res = connection.psql_command(cmd=copy)
    if not res:
        return

    print(f"Collection embedding for '{collection}' exported to '{csv_file_embedding}'")


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print("Usage: python export_collection.py <folder_path> <collection_name>")
        sys.exit(1)
    load_dotenv()
    export_collection_data(
        folder_path=sys.argv[1],
        collection=sys.argv[2],
    )



"""Language detection utilities"""
import spacy
from spacy.language import Language
import langcodes


class Detector():
    """Language detection class"""
    nlp: Language

    def __init__(self):
        """Init internal nlp"""
        try:
            import spacy_fastlang  # noqa: F401 pylint: disable=import-outside-toplevel

            self.nlp = spacy.load('en_core_web_sm')
            self.nlp.add_pipe('language_detector')

        except ImportError as exc:
            raise ImportError(
                "SpaCy `en_core_web_sm` not installed!"
            ) from exc

    def detect(self, phrase: str) -> str:
        """Detect language of generic phrase"""
        doc = self.nlp(phrase)

        return langcodes.Language.make(language=doc._.language).display_name()
